/*! \file tcp.c \brief Stellt die TCP/IP Funkionalitaet bereit */
//***************************************************************************
//*            tcp.c
//*
//*  Sat Jun  3 23:01:42 2006
//*  Copyright  2006  User
//*  Email
//****************************************************************************/
///	\ingroup network
///	\defgroup TCP Der TCP Stack fuer Mikrocontroller (tcp.c)
///	\code #include "tcp.h" \endcode
///	\par Uebersicht
///		Der TCP-Stack fuer Mikrocontroller. Behandelt komplett den TCP-Stack
/// mit Verbindungsaufbau, Abbau und halten. Es werden Ereignisse wie bei Timeouts
/// oder Retransmisions behandelt.	
//****************************************************************************/
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
//@{
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdio.h>

#include "../math/math.h"
#include "../math/checksum.h"
#include "../hardware/timer1.h"

#include "ethernet.h"
#include "arp.h"
#include "ip.h"
#include "tcp.h"

// #define _DEBUG_

#ifdef _DEBUG_
	#include <stdio.h>
	#include "../hardware/uart.h"
#endif

unsigned int TXErrorCounter = 0 ;
unsigned int RXErrorCounter = 0 ;

struct TCP_SOCKET TCP_sockettable[MAX_TCP_CONNECTIONS];

struct TCP_PORT TCP_porttable[MAX_LISTEN_PORTS];
	
void tcp_init( void )
{
		timer1_sec_RegisterCallbackFunction( TCPTimeOutHandler );	
}
	
/*------------------------------------------------------------------------------------------------------------*/
/*!\brief Die TCP-Funktion die aufgerufen wird wenn ein Packet eintrifft.
 * Hier findet die Bearbeitung der eintreffenden Packete statt. Das Packet wird einer Verbindung zugeordnet 
 * oder einen offenen Port wenn die in die TCP_PORT Liste eingetragen ist.
 * Danach wird es je nach Flag bearbeitet.
 * \code #include "test.h"
 * void main( void)
 * \endcode
 * \param 		packet_lenght	Groesse des Packetes.
 * \param 		ethernetbuffer  Zeiger auf den Ethernetbuffer, dieser enthaelt noch die kompletten Header aller Layer.
 * \retval		NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void tcp( unsigned int packet_lenght, unsigned char * ethernetbuffer)
	{
		unsigned int i;
		unsigned int Socket;
		
		struct ETH_header *ETH_packet; 		// ETH_struct anlegen
		ETH_packet = (struct ETH_header *) ethernetbuffer;
		struct IP_header *IP_packet;		// IP_struct anlegen
		IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
			struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];
		
		Socket = GetSocket ( ethernetbuffer );
		
		if ( Socket == 0xffff )
		{
			if ( ( CheckPortInList( ChangeEndian16bit ( TCP_packet->TCP_DestinationPort ) ) == 0xffff ) ) 
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P(PSTR("Verbindung zu unerlaubten Port\n\r"));
				#endif
				/* Wieder auskommendtieren wenn ein Antwort gegeben werden soll
				Socket = RegisterSocket( ethernetbuffer );
				if ( Socket == 0xffff ) return; // Kein Socket Frei, nix machen
				MakeTCPheader( Socket, TCP_RST_FLAG , 0, 0, ethernetbuffer ); // Reset-Flag setzen, TCP bauen
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_NOT_USE; // Connection State wieder auf frei setzen
				*/
				return;
			}
		}
		
		if ( TCP_packet->TCP_ControllFlags == TCP_SYN_FLAG ) // Packet mit SYN ?
			{
				Socket = RegisterSocket( ethernetbuffer ) ;
				if ( Socket != 0xffff ) // Verbinung einem neuem Socket zuordnen
				{
					#ifdef _DEBUG_
						printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
						printf_P(PSTR("Kommende Verbindung auf Port %d [SYN] auf Socket %d\r\n") , ChangeEndian16bit ( TCP_sockettable[ Socket ].DestinationPort ), Socket );
					#endif
					TCP_sockettable[ Socket ].ConnectionState = SOCKET_SYNINIT ; // SYNINIT setzen so lange wie das packet noch nicht beantwortet wurde, wird in MakeTCP gebraucht um de MSS zu senden
					TCP_sockettable[ Socket ].Windowsize = TCP_packet->TCP_Window; // Windowsize setzen, wird gebraucht um zu wissen wiviel gesendet werden kann ohne ACK
					MakeTCPheader( Socket, TCP_SYN_FLAG | TCP_ACK_FLAG, 0 , MAX_RECIVEBUFFER_LENGHT , ethernetbuffer ); // Baue mal den TCP-Header mit Berechnung des Pseudoheader
					TCP_sockettable[ Socket ].SequenceNumber++; // SequenceNumber um 1 erhöhen, das gehört zur SYN-sequence dazu
					TCP_sockettable[ Socket ].ConnectionState = SOCKET_WAIT2SYNACK ; // State für den Socket auf WAIT2SYNACK und den den SYN abschließen zu können
					TCP_sockettable[ Socket ].SendState = SOCKET_READY2SEND ; // bereit zum senden
					TCP_sockettable[ Socket ].Timeoutcounter = 2;
					#ifdef _DEBUG_
						printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
						printf_P(PSTR("Verbindungsanforderung bestaetigt [SYN + ACK]\r\n") , ChangeEndian16bit ( TCP_sockettable[ Socket ].DestinationPort ) );
					#endif
				}
				return;
			}
		
		// Wenn immer noch keine Verbindung zugeordnet beenden
		if ( Socket == 0xffff ) return;
		// Timeout füer Socket zuruecksetzen
		#ifdef _DEBUG_
			printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
			printf_P(PSTR("Verbindungszuordnung auf Socket %d\r\n") , Socket );
		#endif
		// den Timeoutcounter wieder zuruecksetzen für die Verbindung auf den zugeordneten Socket
		TCP_sockettable[ Socket ].Timeoutcounter = TimeOutCounter;
		
		if ( TCP_packet->TCP_ControllFlags == ( TCP_SYN_FLAG + TCP_ACK_FLAG ) ) // Packet mit SYN + ACK ?
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf("Verbindungsbestaetigung [SYN + ACK]\r\n");
				#endif
				// SequenceNumber um 1 erhöhen, das gehört zur SYN-sequence dazu
				TCP_sockettable[ Socket ].SequenceNumber = ChangeEndian32bit( TCP_packet->TCP_AcknowledgeNumber );
				TCP_sockettable[ Socket ].AcknowledgeNumber = ChangeEndian32bit( TCP_packet->TCP_SequenceNumber );
				TCP_sockettable[ Socket ].AcknowledgeNumber++;
				// Windowsize setzen, wird gebraucht um zu wissen wiviel gesendet werden kann ohne ACK
				TCP_sockettable[ Socket ].Windowsize = TCP_packet->TCP_Window; 
				// Baue mal den TCP-Header mit Berechnung des Pseudoheader
				// MakeTCPheader( Socket, TCP_ACK_FLAG, 0 , MAX_RECIVEBUFFER_LENGHT , ethernetbuffer );
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_READY ; // State für den Socket auf WAIT2SYNACK und den den SYN abschließen zu können
				TCP_sockettable[ Socket ].SendState = SOCKET_READY2SEND ; // bereit zum senden
				return;
			}

		// anzahl der Daten im TCP-Datagram berechnen
			i =	( ChangeEndian16bit ( IP_packet->IP_Totallenght )
			- ( ( IP_packet->IP_Version_Headerlen & 0x0f ) * 4
			+ ( ( TCP_packet->TCP_DataOffset & 0xf0 ) >> 2 ) ) ) ;
	
		// Wenn Daten vorhanden einffach mal kopieren in den Socket-Puffer und bestätigung senden
		if ( i != 0 )
		{
			// Richtige Reihnfolger der Daten ?
			// Hier kann erzwungen werden das die in der richtigen reihnfolge ankommen, 
			// wenn nicht wird ein ACK-Packet mit einer alten Seqnummer gesendet
			if ( ( TCP_sockettable[ Socket ].AcknowledgeNumber + i ) == ( ChangeEndian32bit ( TCP_packet->TCP_SequenceNumber ) + i ) )
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("Richtige reinfolge der Packete!\r\n") );
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P(PSTR("%d Byte befinden sich im Recivebuffer [PSH + ACK] %d Byte empfangen\r\n"), TCP_sockettable[ Socket ].Bufferlenght , i);
				#endif
				// Wenn kopieren in Puffer ok war, Acknummer richtig setzen, sonst nix machen, dann wird automatisch die alte Seqnummer gesendet
				// mit der richtigen Windowsize
				if ( CopyTCPdata2socketbuffer( Socket, i , ethernetbuffer ) != 0xffff )
					TCP_sockettable[ Socket ].AcknowledgeNumber = ChangeEndian32bit ( TCP_packet->TCP_SequenceNumber ) + i ;					
			}
			else
			{
				// Naja, falsche reihnfolge kann man nur sagen :-)
				RXErrorCounter++;
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("Falsche reinfolge der Packete!\r\n") );
				#endif
			}
			// ACK senden
			MakeTCPheader( Socket, TCP_ACK_FLAG, 0 , ( MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght ) , ethernetbuffer );
			#ifdef _DEBUG_
				printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
				printf_P( PSTR("Bestaetigung gesendet [ACK]\r\n"), TCP_sockettable[ Socket ].Bufferlenght , i);
			#endif
		}

		if ( TCP_packet->TCP_ControllFlags == ( TCP_PSH_FLAG | TCP_ACK_FLAG ) )		// ACK + PSH empfangen ?
		{
			// naja, ist ja schon alles vorher passiert :-)
			return;
		}	
								
		// Anforderung für das schliessen des Socket/Verbindung
		if ( ( TCP_packet->TCP_ControllFlags == ( TCP_FIN_FLAG | TCP_ACK_FLAG ) ) || ( TCP_packet->TCP_ControllFlags == ( TCP_PSH_FLAG | TCP_FIN_FLAG | TCP_ACK_FLAG ) ) )
		{
			// haben wir die Schliessung angefordert ?
			if ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_WAIT2FINACK )
			{
				TCP_sockettable[ Socket ].AcknowledgeNumber = ChangeEndian32bit ( TCP_packet->TCP_SequenceNumber ) + 1;
				TCP_sockettable[ Socket ].SequenceNumber = ChangeEndian32bit ( TCP_packet->TCP_AcknowledgeNumber );
				MakeTCPheader( Socket, TCP_ACK_FLAG, 0 , ( MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght ) , ethernetbuffer );
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_NOT_USE ;
			}

			// Client hat Schliessung angefordert
			if ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_READY ) 
			{
				TCP_sockettable[ Socket ].SequenceNumber = ChangeEndian32bit ( TCP_packet->TCP_AcknowledgeNumber );
				MakeTCPheader( Socket, TCP_FIN_FLAG | TCP_ACK_FLAG, 0 , ( MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght ) , ethernetbuffer );
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_WAIT2FINACK ;
			}				
			// Laut RFC ist Seqnummer um 1 zu erhöhen beim FIN
			TCP_sockettable[ Socket ].SequenceNumber++;
			#ifdef _DEBUG_
				printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
				printf_P(PSTR("Verbindung wird geschlossen [FIN + ACK]\r\n"), TCP_sockettable[ Socket ].Bufferlenght , i);
			#endif
			return;
		}

		// normale bestätigung der gesendeten Daten
		if ( TCP_packet->TCP_ControllFlags == TCP_ACK_FLAG ) 						// mach mal wenn ACK empfangen
			{
				// War das ein ACK von einen Verbindungsaufbau ? Wenn ja steht die verbindung jetzt
				if ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_WAIT2SYNACK )  			// Wartet socket auf SYN + ACK ?
				{
					TCP_sockettable[ Socket ].ConnectionState = SOCKET_READY2USE;					// Socket auf Ready2Use setzen
					TCP_sockettable[ Socket ].SendState = SOCKET_READY2SEND;						// Socket auf Bereit um senden setzen
					#ifdef _DEBUG_
						printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
						printf_P(PSTR("Verbindung ist aufgebaut und READY2USE [ACK]\r\n"));
					#endif
				}				

				// Waren keine Daten enthalten ?
				if ( i == 0 )
				{
					if ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_WAIT2FIN ) 
					{
						TCP_sockettable[ Socket ].ConnectionState=SOCKET_NOT_USE;
						#ifdef _DEBUG_
							printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
							printf_P( PSTR("Verbindung geschlossen [ACK]\r\n"));
						#endif
					}
					else
					{
						if ( TCP_sockettable[ Socket ].SequenceNumber == ChangeEndian32bit ( TCP_packet->TCP_AcknowledgeNumber ) )
						{
							#ifdef _DEBUG_
								printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
								printf_P( PSTR("Empfang bestaetig [ACK] SendState aud READY2SEND\r\n"));
							#endif
							TCP_sockettable[ Socket ].SendState = SOCKET_READY2SEND;
							TCP_sockettable[ Socket ].SendetBytes = 0;
						}
						else
						{
							TCP_sockettable[ Socket ].SendetBytes = TCP_sockettable[ Socket ].SequenceNumber - ChangeEndian32bit ( TCP_packet->TCP_AcknowledgeNumber );

							#ifdef _DEBUG_
								printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
							#endif

							if ( TCP_packet->TCP_Window > ( TCP_sockettable[ Socket ].Windowsize / 2 ) )
							{
								TCP_sockettable[ Socket ].SendState = SOCKET_DATASENDED ;
							#ifdef _DEBUG_
								printf_P( PSTR("Bestaetigung nicht vollstaendig (%d Byte) [ACK] SendState auf SOCKET_DATASENDED\r\n"), TCP_sockettable[ Socket ].SendetBytes );
							#endif
							}
							else 
							{
								TCP_sockettable[ Socket ].SendState = SOCKET_BUSY ;
							#ifdef _DEBUG_
								printf_P( PSTR("Bestaetigung nicht vollstaendig (%d Byte) [ACK] SendState weiter SOCKET_BUSY\r\n"), TCP_sockettable[ Socket ].SendetBytes );
							#endif
							}

						}
					}
				}
			return;
			}

		return;
	}
		
/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Der TCP-Timeouthandler
 * Dieser Handler schaut zyklisch in alle Sockets und verringert den Timeoutwert. Wenn ein Timeoutwert 0 erreicht
 * hat, wird das Socket geschlossen und der Connectionstats richtig gesetz.
 * \warning Die Funktion wir in der tcp_init() beim timerinterrupt registiert und fortan zyklisch aufgerufen.
 * \param	NONE
 * \retval	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/
void TCPTimeOutHandler( void )
	{
		unsigned int Socket;
		// durch läufe MAX_TCP_CONNECTIONS
		for ( Socket = 0 ; Socket < MAX_TCP_CONNECTIONS ; Socket ++ )
			{
				// Wenn TimeOutcounter unsgleich 0 dann verringern
				if ( TCP_sockettable[ Socket ].Timeoutcounter != 0 ) TCP_sockettable[ Socket ].Timeoutcounter--;
				
				if ( TCP_sockettable[ Socket ].Timeoutcounter == 0 && TCP_sockettable[ Socket ].ConnectionState != SOCKET_NOT_USE )
				{
					unsigned char * ethernetbuffer;
					ethernetbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT );
					cli();
					MakeTCPheader( Socket, TCP_RST_FLAG | TCP_FIN_FLAG | TCP_ACK_FLAG , 0 , 0 , ethernetbuffer );
					TCP_sockettable[ Socket ].ConnectionState = SOCKET_NOT_USE ;
					sei();
					#ifdef _DEBUG_
						printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
						printf_P( PSTR("Connection timeout oder Verbindungsabriss\r\n"));
					#endif
				}
			}
	}

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Holt das naechste freie Socket
 * \warning Die Funktion wir in der tcp_init() beim timerinterrupt registiert und fortan zyklisch aufgerufen.
 * \param	NONE
 * \retval	SOCKET	Gibt die Nummer des naechsten Freien SOcket.
 */
/*------------------------------------------------------------------------------------------------------------*/
unsigned int Getfreesocket( void )
	{
		unsigned int Socket;
		for ( Socket = 0 ; Socket < MAX_TCP_CONNECTIONS ; Socket++ )
			{
				if ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_NOT_USE ) return( Socket );
			}
		return(0xffff);
	}		

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Ordnet das Packet einen Socket zu.
 * \param	ethernetbuffer	Zeiger auf den Ethernetpuffer.
 * \retval	Socket	Im Erfolgsfall die Socketnummer, im Fehlerfall 0xffff
 */
/*------------------------------------------------------------------------------------------------------------*/
unsigned int GetSocket( unsigned char * ethernetbuffer )
	{
		struct ETH_header *ETH_packet; 		// ETH_struct anlegen
		ETH_packet = (struct ETH_header *) ethernetbuffer;
		struct IP_header *IP_packet;		// IP_struct anlegen
		IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];

		unsigned int Socket;
		
		for ( Socket = 0 ; Socket < MAX_TCP_CONNECTIONS ; Socket++ ) 
			{
				if ( 	TCP_sockettable[ Socket ].ConnectionState != SOCKET_NOT_USE 
						&& TCP_sockettable[ Socket ].SourcePort == TCP_packet->TCP_SourcePort 
						&& TCP_sockettable[ Socket ].DestinationPort == TCP_packet->TCP_DestinationPort 
						&& TCP_sockettable[ Socket ].SourceIP == IP_packet->IP_SourceIP ) return( Socket );
			}
		return(0xffff);
	}		
	
/* -----------------------------------------------------------------------------------------------------------
Reistriert einen neuen Socket wenn noch einer frei ist und sichert alle nötigen daten in ihn
------------------------------------------------------------------------------------------------------------*/
unsigned int RegisterSocket( unsigned char *ethernetbuffer)
	{
		struct ETH_header *ETH_packet; 		// ETH_struct anlegen
		ETH_packet = (struct ETH_header *) ethernetbuffer;
		struct IP_header *IP_packet;		// IP_struct anlegen
		IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];

		unsigned int Socket;
		
		Socket = Getfreesocket();
		if ( Socket == 0xffff ) return(0xffff);
		
		// Register den SOCKET
		TCP_sockettable[ Socket ].SourcePort = TCP_packet->TCP_SourcePort;
		TCP_sockettable[ Socket ].DestinationPort = TCP_packet->TCP_DestinationPort;
		TCP_sockettable[ Socket ].SourceIP = IP_packet->IP_SourceIP;
		TCP_sockettable[ Socket ].SequenceNumber =~ ChangeEndian32bit( TCP_packet->TCP_SequenceNumber );
		TCP_sockettable[ Socket ].AcknowledgeNumber = ChangeEndian32bit( TCP_packet->TCP_SequenceNumber ) + 1;
		TCP_sockettable[ Socket ].SendState = SOCKET_READY2SEND;
		TCP_sockettable[ Socket ].Timeoutcounter = 10;
		TCP_sockettable[ Socket ].Bufferlenght = 0;
		TCP_sockettable[ Socket ].Windowsize = TCP_packet->TCP_Window ;
		TCP_sockettable[ Socket ].SendetBytes = 0;
		
		for ( unsigned int i = 0 ; i < 6 ; i++ ) TCP_sockettable[ Socket ].MACadress[i] = ETH_packet->ETH_sourceMac[i];
		
		return( Socket );
	}

/* -----------------------------------------------------------------------------------------------------------
Baut einen TCP-header und berechnet den Pseudoheader und die Checksumme 
	Übergeben werden müssen die TCP_FLAGS, Datenlänge den Datensegments, die Windowsize, und der Pointer auf den Buffer
------------------------------------------------------------------------------------------------------------*/	
void MakeTCPheader( unsigned int Socket, unsigned char TCP_flags, unsigned int Datalenght, unsigned int Windowsize, unsigned char * ethernetbuffer )
	{
		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT ];

		struct IP_Pseudoheader *IP_pseudopacket;
		IP_pseudopacket = ( struct IP_Pseudoheader *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT - IP_PSEUDOHEADER_LENGHT ];

		// Bastel mal den Pseudoheader
		IP_pseudopacket->IP_SourceIP = TCP_sockettable[ Socket ].SourceIP;
		IP_pseudopacket->IP_DestinationIP = myIP;
		IP_pseudopacket->IP_ZERO = 0x0;
		IP_pseudopacket->IP_Protokoll = 0x6;
		IP_pseudopacket->IP_TCP_lenght = ChangeEndian16bit( TCP_HEADER_LENGHT + Datalenght );										
			
		// TCP_header basteln
		TCP_packet->TCP_SourcePort = TCP_sockettable[ Socket ].DestinationPort;
		TCP_packet->TCP_DestinationPort = TCP_sockettable[ Socket ].SourcePort;
		TCP_packet->TCP_SequenceNumber = ChangeEndian32bit( TCP_sockettable[ Socket ].SequenceNumber );
		TCP_packet->TCP_AcknowledgeNumber = ChangeEndian32bit( TCP_sockettable[ Socket ].AcknowledgeNumber );
		TCP_packet->TCP_DataOffset = ( ( TCP_HEADER_LENGHT << 2 ) & 0xf0 ) ;
		TCP_packet->TCP_ControllFlags = TCP_flags ;
		TCP_packet->TCP_Window = ChangeEndian16bit( Windowsize );
		TCP_packet->TCP_Checksum = 0x0;
		TCP_packet->TCP_UrgentPointer = 0x0;
		
		if ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_SYNINIT )
		{
			// beschreibe das Optionfld mit MSS = MAX_RECIVEBUFFER_LENGHT wenn der SYN ausgehandelt wird
			TCP_packet->TCP_Options[0] = 0x02;
			TCP_packet->TCP_Options[1] = 0x04;
			TCP_packet->TCP_Options[2] = ( MAX_RECIVEBUFFER_LENGHT >> 8 ) & 0x00ff ;
			TCP_packet->TCP_Options[3] = MAX_RECIVEBUFFER_LENGHT & 0x00ff ;
		}
		else
		{
			// lasse Optionefeld leer wenn kein SYN
			TCP_packet->TCP_Options[0] = 0;
			TCP_packet->TCP_Options[1] = 0;
			TCP_packet->TCP_Options[2] = 0;
			TCP_packet->TCP_Options[3] = 0;
		}
			
		TCP_packet->TCP_Checksum = ChangeEndian16bit( Checksum_16( 	&ethernetbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT - IP_PSEUDOHEADER_LENGHT ],
																	TCP_HEADER_LENGHT + IP_PSEUDOHEADER_LENGHT  + Datalenght ) ) ;
		MakeIPheader( TCP_sockettable[ Socket ].SourceIP, PROTO_TCP, TCP_HEADER_LENGHT + Datalenght , ethernetbuffer );
		MakeETHheader( TCP_sockettable[ Socket ].MACadress, ethernetbuffer );
		sendEthernetframe( ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT + Datalenght, ethernetbuffer );
		
		return;
	}
		
/* -----------------------------------------------------------------------------------------------------------
Kopiert die Datenm die mit TCP-übertragen wurden in den Socketbuffer
------------------------------------------------------------------------------------------------------------*/
unsigned int CopyTCPdata2socketbuffer( unsigned int Socket, unsigned int Datalenght , unsigned char *ethernetbuffer )
	{

		struct ETH_header *ETH_packet; 		// ETH_struct anlegen
		ETH_packet = (struct ETH_header *) ethernetbuffer;
		struct IP_header *IP_packet;		// IP_struct anlegen
		IP_packet = ( struct IP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH];
		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + ((IP_packet->IP_Version_Headerlen & 0x0f) * 4 )];
			
		// Abbrechen wenn die Datem nicht mehr in den buffer passen
		if ( ( MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght ) < Datalenght ) return( 0xffff );
		
		unsigned int Offset = ETHERNET_HEADER_LENGTH + ( IP_packet->IP_Version_Headerlen & 0x0f ) * 4 + ( ( TCP_packet->TCP_DataOffset & 0xf0 ) >> 2 ) ;
				
		unsigned int i;
		
		for (  i = 0 ; i < Datalenght ; i++ )
		{
				TCP_sockettable[ Socket ].Recivebuffer[ i + TCP_sockettable[ Socket ].Bufferlenght ] = ethernetbuffer[ Offset + i ];
		}
		TCP_sockettable[ Socket ].Bufferlenght = TCP_sockettable[ Socket ].Bufferlenght + Datalenght ;
					
		return( 0 );
	}		
	
/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Registriert einen Port auf den gelauscht wird für einegehende Verbindungen
 * \param	Port	Der Port auf den Gelauscht werden soll
 * \retval	Im Erfolgsfall 0, im Fehlerfall 0xffff
 */
/*------------------------------------------------------------------------------------------------------------*/
unsigned int RegisterTCPPort( unsigned int Port )
	{
		unsigned int i;
		
		for ( i = 0 ; i < MAX_LISTEN_PORTS ; i ++ ) 
			{
				if ( TCP_porttable[i].TCP_Port == TCP_Port_not_use ) 
				{
					TCP_porttable[i].TCP_Port = Port;
					return(0);
				}
			}
		return(0xffff);
	}

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Traegt einen Port aus der Liste der Port aus auf denen gelauscht werden soll
 * \param	Port	Der Port auf den Gelauscht werden soll
 * \return	Im Erfolgsfall 0, im Fehlerfall 0xffff
 */
/*------------------------------------------------------------------------------------------------------------*/
void UnRegisterTCPPort( unsigned int Port )
	{
		unsigned int i;
		
		for ( i = 0 ; i < MAX_LISTEN_PORTS ; i ++ ) 
			{
				if ( TCP_porttable[i].TCP_Port == Port ) TCP_porttable[i].TCP_Port = TCP_Port_not_use;
			}
		return;
	}
	
/* -----------------------------------------------------------------------------------------------------------
Schaut ob ein Port schon in der Liste ist
------------------------------------------------------------------------------------------------------------*/
unsigned int CheckPortInList( unsigned int Port )
	{
		unsigned int i;
		
		for ( i = 0 ; i < MAX_LISTEN_PORTS ; i ++ ) 
			{
				if ( TCP_porttable[i].TCP_Port == Port ) return(0);
			}
		return( 0xffff );
	}
	
/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Schaut ob auf einen Port ein Verbindung eingegangen ist.
 * \param	Port	Der Port der abgefragt wird.
 * \return	Im Erfolgsfall die Socketnummer, im Fehlerfall 0xffff
 */
/*------------------------------------------------------------------------------------------------------------*/	
unsigned int CheckPortRequest( unsigned int Port )
	{
		for ( unsigned int i = 0 ; i < MAX_TCP_CONNECTIONS ; i ++ ) 
			{
				if ( TCP_sockettable[i].DestinationPort == ChangeEndian16bit( Port ) && TCP_sockettable[i].ConnectionState == SOCKET_READY2USE ) 
				{
					TCP_sockettable[i].ConnectionState = SOCKET_READY;
					return(i);
				}
			}
		return( 0xffff );
	}

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Gibt den Status des Socket aus.
 * \param	Socket		Die Socketnummer vom dem der Status zurueckgegeben werden soll.
 * \return	Der Socketstate
 */
/*------------------------------------------------------------------------------------------------------------*/	
unsigned int CheckSocketState( unsigned int Socket )
	{
		return( TCP_sockettable[ Socket ].ConnectionState );
	}

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Schliesst ein Socket und Beendet die TCP-Verbindung.
 * \param	Socket		Die Socketnummer die geschlossen werden soll.
 * \return	NONO
 */
/*------------------------------------------------------------------------------------------------------------*/		
void CloseTCPSocket( unsigned int Socket)
	{
		if ( Socket >= MAX_TCP_CONNECTIONS || TCP_sockettable[Socket].ConnectionState == SOCKET_NOT_USE ) return( NO_SOCKET_USED );
					
		unsigned char timer;
		
		timer = timer1_msec_RegisterCoundowntimer();
		timer1_msec_SetCountdownTimer( timer , CLOSETIMEOUT );

		while ( 1 )
		{
			if ( ( TCP_sockettable[ Socket ].SendState == SOCKET_READY2SEND ) && ( timer1_msec_GetCountdownTimer( timer ) != 0 ) )
			{
				timer1_msec_ReleaseCountdownTimer( timer );
				break;
			}
			if ( timer1_msec_GetCountdownTimer( timer ) == 0 ) 
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("Close-Timeout. Verbindung wird geschlossen, Socket nicht bereit!\r\n") );
				#endif
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_NOT_USE ;
				timer1_msec_ReleaseCountdownTimer( timer );
				return;
			}
		}

		#ifdef _DEBUG_
			printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
			printf_P( PSTR("Verbindung kann geschlossen werden\r\n") );
		#endif
		
		LockEthernet();
		
		unsigned char * ethernetbuffer;
		ethernetbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT );
		
		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT ];

		MakeTCPheader( Socket, TCP_FIN_FLAG | TCP_ACK_FLAG , 0, 0, ethernetbuffer );

		TCP_sockettable[ Socket ].ConnectionState = SOCKET_WAIT2FINACK;
	
		FreeEthernet();		

		timer = timer1_msec_RegisterCoundowntimer();
		timer1_msec_SetCountdownTimer( timer , CLOSETIMEOUT );
		
		while ( 1 )
		{
			if ( ( TCP_sockettable[ Socket ].ConnectionState == SOCKET_NOT_USE ) && ( timer1_msec_GetCountdownTimer( timer ) != 0 ) )
			{
				timer1_msec_ReleaseCountdownTimer( timer );
				break;
			}
			if ( timer1_msec_GetCountdownTimer( timer ) == 0 ) 
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("Close-Timeout. Verbindung wird geschlossen. State von Socket %d auf SOCKET_NOT_USE\r\n"), Socket );
				#endif

				LockEthernet();
				MakeTCPheader( Socket, TCP_RST_FLAG , 0, 0, ethernetbuffer );
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_NOT_USE ;
				FreeEthernet();

				timer1_msec_ReleaseCountdownTimer( timer );
				break;
			}
		}
		return;
	}	

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Sendet Daten ueber ein Socket aus dem RAM/FLASH/EEPROM.
 * \param	Socket		Die Socketnummer die zum versnden benutzt werden soll.
 * \param	Datalenght	Die Datenlaenge in Bytes die sersendet werden soll.
 * \param	Sendbuffer	Zeiger auf die Dten im RAM der versendet werden soll.
 * \param	Mode		Art des Zeigers (RAM/FLASH/EEPROM).
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/	
void SendData_RPE( unsigned int Socket, unsigned int Datalenght, unsigned char * Sendbuffer, unsigned char Mode )	
	{
		if ( Socket >= MAX_TCP_CONNECTIONS || TCP_sockettable[Socket].ConnectionState == SOCKET_NOT_USE ) return( NO_SOCKET_USED );
		
		unsigned char timer;
		
		timer = timer1_msec_RegisterCoundowntimer();
		timer1_msec_SetCountdownTimer( timer , CLOSETIMEOUT );

		while ( 1 )
		{
			if ( ( ( TCP_sockettable[ Socket ].SendState == SOCKET_READY2SEND ) || ( TCP_sockettable[Socket].SendState == SOCKET_DATASENDED ) ) && ( timer1_msec_GetCountdownTimer( timer ) != 0 ) )
			{
				timer1_msec_ReleaseCountdownTimer( timer );
				break;
			}
			if ( timer1_msec_GetCountdownTimer( timer ) == 0 ) 
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("Verbindung nicht frei. Daten nicht gesendet\r\n") );
				#endif
				timer1_msec_ReleaseCountdownTimer( timer );
				return;
			}
		}
		
		LockEthernet();

		unsigned char * ethernetbuffer;
		ethernetbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT + Datalenght );

		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) &ethernetbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT ];	

		unsigned int Offset = ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT ;
		unsigned int i;

		switch ( Mode )
		{
			case RAM:	for ( i = 0 ; i < Datalenght ; i++ )
								ethernetbuffer[ Offset + i ] = Sendbuffer[ i ]; // kopiere die daten in den Sendebiffer
						break;
			case FLASH:	for ( i = 0 ; i < Datalenght ; i++ )
								ethernetbuffer[ Offset + i ] = pgm_read_byte ( Sendbuffer + i ) ;
						break;
			default:	return( 0xffff );
		}

		MakeTCPheader( Socket, TCP_PSH_FLAG | TCP_ACK_FLAG , Datalenght, MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght , ethernetbuffer );
		TCP_sockettable[ Socket ].SequenceNumber = TCP_sockettable[ Socket ].SequenceNumber + i ;
		TCP_sockettable[ Socket ].SendetBytes = TCP_sockettable[ Socket ].SendetBytes + Datalenght ;
		
		if ( ( TCP_sockettable[ Socket ].Windowsize - TCP_sockettable[ Socket ].SendetBytes ) > MAX_TCP_Datalenght ) TCP_sockettable[ Socket ].SendState = SOCKET_DATASENDED ;
		else TCP_sockettable[ Socket ].SendState = SOCKET_BUSY ;
		
		#ifdef _DEBUG_
			printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
			printf_P( PSTR("%d Bytes gesendet [PSH + ACK] und SendState auf BUSY\r\n"), i );
		#endif

		FreeEthernet();
 
		return ;
	}

/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Sendet Daten ueber ein Socket aus dem FLASH.
 * \param	Socket		Die Socketnummer die zum versnden benutzt werden soll.
 * \param	Datalenght	Die Datenlaenge in Bytes die sersendet werden soll.
 * \param	Sendbuffer	Zeiger auf die Daten im FLASH der versendet werden soll.
 * \param	Mode		Wie der Zeiger interpretiert werden soll, als RAM/FLASH/EEPROM.
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/	
unsigned int PutSocketData_RPE( unsigned int Socket, unsigned int Datalenght, unsigned char * Sendbuffer, unsigned char Mode )
	{
		// schau mal ob socket gültig ist
		if ( Socket >= MAX_TCP_CONNECTIONS || TCP_sockettable[Socket].ConnectionState == SOCKET_NOT_USE ) return( NO_SOCKET_USED );
			
		unsigned int i;
		unsigned long SequenceNumber;
		unsigned char timer;
		unsigned char TransmissionCounter = 0 ;

		// seqnumber vor dem senden speichern + daten länge
		// damit wird verglichen ob alles daten quitiert worden sind, und es zu fehler bei der übertragung kam
		SequenceNumber = TCP_sockettable[Socket].SequenceNumber + Datalenght ;
					
		// reseviere mal einen timer
		timer = timer1_msec_RegisterCoundowntimer();

		while ( 1 )
		{
			// ist das socket noch offen ?
			if ( TCP_sockettable[Socket].ConnectionState == SOCKET_NOT_USE || TCP_sockettable[Socket].ConnectionState == SOCKET_CLOSE )
			{
				timer1_msec_ReleaseCountdownTimer( timer );
				return( SOCKET_NOT_USE );
			}	
			
			// beim ersten durchlauf senden, denn timer steht noch auf null
			if ( timer1_msec_GetCountdownTimer( timer ) == 0 )
			{
				// wenn retransmission stattgefunden hat, seqnumber wird reseten
				if ( TransmissionCounter != 0 ) TCP_sockettable[Socket].SequenceNumber = SequenceNumber - Datalenght;
				i = 0;
				// daten senden
				while ( ( Datalenght - i ) > MAX_TCP_Datalenght )
				{
					switch ( Mode )
					{
						case RAM:	SendData_RPE( Socket, MAX_TCP_Datalenght, &Sendbuffer[i], RAM );
									break;
						case FLASH:	SendData_RPE( Socket, MAX_TCP_Datalenght, &Sendbuffer[i], FLASH );
									break;
						default:	return( 0xffff );
					}
					i = i + MAX_TCP_Datalenght;
				}

				if ( Datalenght != i )
				{
					switch ( Mode )
					{
						case RAM:	SendData_RPE( Socket , Datalenght - i , &Sendbuffer[i], RAM );
									break;
						case FLASH:	SendData_RPE( Socket , Datalenght - i , &Sendbuffer[i], FLASH );
									break;
						default:	return( 0xffff );
					}
				}
				// anzahl der versuche erhöhen
				#ifdef _DEBUG_
					if ( TransmissionCounter != 0 ) 
					{
						printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
						printf_P( PSTR("Timeout. Daten nochmal senden. %d Versuch\r\n"), TransmissionCounter );
					}
				#endif;
				if ( TransmissionCounter != 0 ) TXErrorCounter++;
				TransmissionCounter++;
				
				// timeouttimer starten
				timer1_msec_SetCountdownTimer( timer , RETRANSMISSIONTIMEOUT * TransmissionCounter );
			}
			
			// schau mal wie oft den schon gesendet worden ist, wenn max erreicht socket schliessen
			if ( TransmissionCounter > MAX_TCP_RETRANSMISSIONS )
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("Timeout. Verbindung wird geschlossen\r\n"));
				#endif
				// timer wirder freigeben
				timer1_msec_ReleaseCountdownTimer( timer );
				// socket frei geben
				CloseTCPSocket( Socket );
				return( NO_SOCKET_USED );
			}
			
			// wenn der empfang bestatigt worden ist vor dem timeout, alles okay
			if ( SequenceNumber == TCP_sockettable[Socket].SequenceNumber )
			{
				timer1_msec_ReleaseCountdownTimer( timer );
				break;
			}
		}
		
		// anzahl der gesendet daten zurück geben
		return( Datalenght );		
	}
	
/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Sendet Daten ueber ein Socket aus dem FLASH.
 * \param	Socket		Die Socketnummer die zum versnden benutzt werden soll.
 * \param	Datalenght	Die Datenlaenge in Bytes die sersendet werden soll.
 * \param	Sendbuffer	Zeiger auf die Daten im FLASH der versendet werden soll.
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/	
unsigned int PutSocketData_P( unsigned int Socket, unsigned int Datalenght, const prog_char * Sendbuffer )
	{
		return( PutSocketData_RPE( Socket, Datalenght, Sendbuffer, FLASH ) );
	}
	
/*-----------------------------------------------------------------------------------------------------------*/
/*!\brief Sendet Daten ueber ein Socket aus dem RAM.
 * \param	Socket		Die Socketnummer die zum versnden benutzt werden soll.
 * \param	Datalenght	Die Datenlaenge in Bytes die sersendet werden soll.
 * \param	Sendbuffer	Zeiger auf die Dten im RAM der versendet werden soll.
 * \return	NONE
 */
/*------------------------------------------------------------------------------------------------------------*/	
unsigned int PutSocketData( unsigned int Socket, unsigned int Datalenght, unsigned char * Sendbuffer )
	{
		return( PutSocketData_RPE( Socket, Datalenght, Sendbuffer, RAM ) );
	}

/* -----------------------------------------------------------------------------------------------------------
Holt daten aus dem Socketbuffer bis zum nächsten '\n' 
Param: Socketnummer, maximale Buffergröße dir gelsen werden darf und der buffer in den geschrieben werden soll
------------------------------------------------------------------------------------------------------------*/
unsigned int GetSocketNextLine( unsigned int Socket , unsigned int bufferlen, unsigned char *buffer)
	{
		if ( Socket >= MAX_TCP_CONNECTIONS || TCP_sockettable[Socket].ConnectionState == SOCKET_NOT_USE ) return( NO_SOCKET_USED );
		if ( TCP_sockettable[Socket].Bufferlenght == 0 ) return(0);
		
		LockEthernet();
		
		unsigned int i,j=0;

		// Buffer kopieren bis \r\n gefunden wurden oder ende erreicht wurden
		for ( i = 0 ; ( i < bufferlen ) && ( i < TCP_sockettable[ Socket ].Bufferlenght ) ; i++ )
		{
			buffer[ i ] = TCP_sockettable[ Socket ].Recivebuffer[ i ];
			if ( ( TCP_sockettable[ Socket ].Recivebuffer[ i ] == '\r' ) && ( TCP_sockettable[ Socket ].Recivebuffer[ i + 1 ] == '\n' ) ) 
			{
				buffer[ i + 1 ] = TCP_sockettable[ Socket ].Recivebuffer[ i + 1 ];
				i = i + 2 ;
				break;
			}
		}
				
		// anzahl der restlichen byte im buffer schreiben
		TCP_sockettable[ Socket ].Bufferlenght = TCP_sockettable[ Socket ].Bufferlenght - i;

		// die verbliebenen Bytes zum anfang des Buffer kopieren
		for ( j = 0 ; j < TCP_sockettable[ Socket ].Bufferlenght ; j++)			
			TCP_sockettable[ Socket ].Recivebuffer[ j ] = TCP_sockettable[ Socket ].Recivebuffer[ j + i ] ;

		#ifdef _DEBUG_
			printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
			printf_P( PSTR("%d Byte aus Recivebuffer augeholt, %d byte noch im Recivebuffer\n\r"), i , TCP_sockettable[ Socket ].Bufferlenght );
		#endif
		
		// sende ein Windowupdate-Paket wenn buffer zu 7/8 frei ist
		if ( TCP_sockettable[ Socket ].Bufferlenght < ( MAX_RECIVEBUFFER_LENGHT / 8 ) )
		{
			unsigned char * ackbuffer;
			ackbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT );

			struct TCP_header *TCP_packet;		// TCP_struct anlegen
			TCP_packet = ( struct TCP_header *)&ackbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT ];							

			MakeTCPheader( Socket, TCP_ACK_FLAG , 0 , MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght , ackbuffer );

			#ifdef _DEBUG_
				printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
				printf_P( PSTR("Neue Windowsize von %d Byte gesendet [ACK]\r\n"), MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght );
			#endif

		}

		FreeEthernet();

		return(i);		
	}

/* -----------------------------------------------------------------------------------------------------------
Holt alle Daten aus den Socketbuffer
------------------------------------------------------------------------------------------------------------*/	
unsigned int GetSocketData( unsigned int Socket , unsigned int bufferlen, unsigned char *buffer)
	{
		if ( Socket >= MAX_TCP_CONNECTIONS || TCP_sockettable[Socket].ConnectionState == SOCKET_NOT_USE ) return( NO_SOCKET_USED );
		if ( TCP_sockettable[ Socket ].Bufferlenght == 0 ) return ( 0 );
		
		LockEthernet();
		
		unsigned int i = 0;
		
		// Wenn der Buffer kleiner als der Empfangsbuffer Buffer voll machen, sonst alles kopieren
		if ( TCP_sockettable[ Socket ].Bufferlenght > bufferlen )
		{
			//memcpy( buffer, TCP_sockettable[ Socket ].Recivebuffer, bufferlen);
			for ( i = 0 ; i < bufferlen ; i++ )
				buffer[i] = TCP_sockettable[ Socket ].Recivebuffer[i];

			TCP_sockettable[ Socket ].Bufferlenght = TCP_sockettable[ Socket ].Bufferlenght - bufferlen ;

			//memcpy( TCP_sockettable[ Socket ].Recivebuffer, &TCP_sockettable[ Socket ].Recivebuffer[ bufferlen ], TCP_sockettable[ Socket ].Bufferlenght );
			for ( i = 0 ; i < TCP_sockettable[ Socket ].Bufferlenght ; i++ )
				TCP_sockettable[ Socket ].Recivebuffer[i] = TCP_sockettable[ Socket ].Recivebuffer[ i + bufferlen ];

			i = bufferlen;
		}
		else
		{
			//memcpy( buffer, TCP_sockettable[ Socket ].Recivebuffer, TCP_sockettable{ Socket ].Bufferlenght );
			for ( i = 0 ; i < TCP_sockettable[ Socket ].Bufferlenght ; i++ )
				buffer[ i ] = TCP_sockettable[ Socket ].Recivebuffer[ i ];

			TCP_sockettable[ Socket ].Bufferlenght = 0;
		}
		
		#ifdef _DEBUG_
			printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
			printf_P( PSTR("%d Byte aus Recivebuffer augeholt, %d byte noch im Recivebuffer\n\r"), i , TCP_sockettable[ Socket ].Bufferlenght );
		#endif
		
		// sende ein Windowupdate-Paket wenn buffer zu 7/8 frei ist
		if ( TCP_sockettable[ Socket ].Bufferlenght < ( MAX_RECIVEBUFFER_LENGHT / 8 ) )
		{
			unsigned char * ackbuffer;
			ackbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT );

			struct TCP_header *TCP_packet;		// TCP_struct anlegen
			TCP_packet = ( struct TCP_header *)&ackbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT ];							

			MakeTCPheader( Socket, TCP_ACK_FLAG , 0 , MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght , ackbuffer );

			#ifdef _DEBUG_
				printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
				printf_P( PSTR("Neue Windowsize von %d Byte gesendet [ACK]\r\n"), MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght );
			#endif
		}

		FreeEthernet();

		return( i );
	}

/* -----------------------------------------------------------------------------------------------------------
Baut eine neue Vrbindung zu einen anderen Rechner auf
------------------------------------------------------------------------------------------------------------*/
unsigned int Connect2IP( unsigned long IP, unsigned int Port )
	{
		unsigned int Socket;
		// hole einen freien SOCKET
		Socket = Getfreesocket();
		// Wenn kein SOCKET mehr frei, EXIT!
		if ( Socket == NO_SOCKET_USED ) return ( NO_SOCKET_USED );

		// Arp-request senden und schaun ob ip vorhanden
		if ( IS_ADDR_IN_MY_SUBNET( IP, Netmask ) )
		{
			if ( IS_BROADCAST_ADDR( IP, Netmask ) ) 
			{
				for( unsigned char i = 0 ; i < 6 ; i++ ) TCP_sockettable[ Socket ].MACadress[i] = 0xff;
			}
			else if ( GetIP2MAC( IP, &TCP_sockettable[ Socket ].MACadress ) == NO_ARP_ANSWER ) return ( NO_SOCKET_USED );
		}
		else if ( GetIP2MAC( Gateway , &TCP_sockettable[ Socket ].MACadress ) == NO_ARP_ANSWER ) return ( NO_SOCKET_USED );
		
		// Socket reservieren
		TCP_sockettable[ Socket ].ConnectionState = SOCKET_SYNINIT;
		TCP_sockettable[ Socket ].DestinationPort =~ ss + ms;
		
		unsigned char * ethernetbuffer;
		ethernetbuffer = (unsigned char*) __builtin_alloca (( size_t ) ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT + TCP_HEADER_LENGHT );

		LockEthernet();
	
		struct TCP_header *TCP_packet;		// TCP_struct anlegen
		TCP_packet = ( struct TCP_header *) ethernetbuffer[ETHERNET_HEADER_LENGTH + IP_HEADER_LENGHT ];
		
		// Register den SOCKET
		TCP_sockettable[ Socket ].SourcePort = ChangeEndian16bit( Port ) ;
		TCP_sockettable[ Socket ].SourceIP = IP;
		TCP_sockettable[ Socket ].SequenceNumber =~ 0x12345678;
		TCP_sockettable[ Socket ].AcknowledgeNumber = 0x0 ;
		TCP_sockettable[ Socket ].SendState = SOCKET_READY2SEND;
		TCP_sockettable[ Socket ].Timeoutcounter = 10;
		TCP_sockettable[ Socket ].Bufferlenght = 0;
		TCP_sockettable[ Socket ].Windowsize = 0;
		TCP_sockettable[ Socket ].SendetBytes = 0;		

		MakeTCPheader( Socket, TCP_SYN_FLAG , 0 , MAX_RECIVEBUFFER_LENGHT - TCP_sockettable[ Socket ].Bufferlenght , ethernetbuffer );
	
		FreeEthernet();

		#ifdef _DEBUG_
			printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
			printf_P( PSTR("Verbindungaufbau angefordert [SYN]\r\n") );
		#endif

		unsigned char timer;
		
		timer = timer1_msec_RegisterCoundowntimer();
		timer1_msec_SetCountdownTimer( timer , CONNECTTIMEOUT );

		while ( 1 )
		{
			if ( (  TCP_sockettable[ Socket ].ConnectionState == SOCKET_READY ) && ( timer1_msec_GetCountdownTimer( timer ) != 0 ) )
			{
				timer1_msec_ReleaseCountdownTimer( timer );
				MakeTCPheader( Socket, TCP_ACK_FLAG, 0 , MAX_RECIVEBUFFER_LENGHT , ethernetbuffer );
				return( Socket );
			}
			if ( timer1_msec_GetCountdownTimer( timer ) == 0 ) 
			{
				#ifdef _DEBUG_
					printf_P( PSTR("%02d:%02d:%02d.%02d: ") ,hh,mm,ss,ms);
					printf_P( PSTR("SYN-Timeout. Verbindung wird geschlossen\r\n") );
				#endif
				TCP_sockettable[ Socket ].ConnectionState = SOCKET_NOT_USE ;
				timer1_msec_ReleaseCountdownTimer( timer );
				return( SOCKET_NOT_USE );
			}
		}
}
